# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import uuid
import logging
from pathlib import Path

# import boto3

# from bentoml.utils.ruamel_yaml import YAML
try:
    from ruamel.yaml import YAML
except ModuleNotFoundError:
    from ruamel_yaml import YAML
# from bentoml.exceptions import (
#     BentoMLException,
#     InvalidArgument,
#     YataiDeploymentException,
# )
# from bentoml.saved_bundle import loader
# from bentoml.utils import status_pb_to_error_code_and_message
# from bentoml.utils.s3 import create_s3_bucket_if_not_exists
# from bentoml.utils.tempdir import TempDirectory
# from bentoml.yatai.deployment.aws_utils import (
#     validate_sam_template,
#     FAILED_CLOUDFORMATION_STACK_STATUS,
#     cleanup_s3_bucket_if_exist,
# )
# from bentoml.yatai.deployment.aws_lambda.utils import (
#     init_sam_project,
#     lambda_deploy,
#     lambda_package,
#     reduce_bundle_size_and_upload_extra_resources_to_s3,
#     total_file_or_directory_size,
#     LAMBDA_FUNCTION_LIMIT,
#     LAMBDA_FUNCTION_MAX_LIMIT,
# )

# from bentoml.yatai.deployment.operator import DeploymentOperatorBase
# from bentoml.yatai.deployment.utils import (
#     raise_if_api_names_not_found_in_bento_service_metadata,
# )
# from bentoml.yatai.deployment.aws_utils import (
#     generate_aws_compatible_string,
#     get_default_aws_region,
#     ensure_sam_available_or_raise,
# )
# from bentoml.yatai.deployment.docker_utils import ensure_docker_available_or_raise
# from bentoml.yatai.proto import status_pb2
# from bentoml.yatai.proto.deployment_pb2 import (
#     DeploymentState,
#     ApplyDeploymentResponse,
#     DeleteDeploymentResponse,
#     DescribeDeploymentResponse,
# )
# from bentoml.yatai.proto.repository_pb2 import GetBentoRequest, BentoUri
# from bentoml.yatai.status import Status


logger = logging.getLogger(__name__)

def create_aws_lambda_cloudformation_template_file(
    project_name,
    project_dir,
    namespace,
    deployment_name,
    api_names,
    s3_bucket_name,
    memory_size,
    timeout,
):
    template_file_path = os.path.join(project_dir, 'template.yaml')
    yaml = YAML()
    sam_config = {
        'AWSTemplateFormatVersion': '2010-09-09',
        'Transform': 'AWS::Serverless-2016-10-31',
        'Globals': {
            'Function': {'Timeout': timeout}
        },
        'Resources': {},
        'Outputs': {
            'S3Bucket': {
                'Value': s3_bucket_name,
                'Description': 'S3 Bucket for saving artifacts and lambda bundle',
            }
        },
    }
    for api_name in api_names:
        sam_config['Resources'][api_name] = {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'PackageType' : "Image",
                'FunctionName': f'{namespace}-{deployment_name}-{api_name}',
                'Timeout': timeout,
                'MemorySize': memory_size,
                'Events': {
                    'Api': {
                        'Type': 'HttpApi',
                        'Properties': {
                            'Path': '/{proxy+}',
                            'Method': 'any',
                        },
                    }
                },
                # 'ImageUri': "639150796852.dkr.ecr.us-east-2.amazonaws.com/cldz-lambda:sam-test",
                'Policies': [{'S3ReadPolicy': {'BucketName': s3_bucket_name}}],
                'Environment': {
                    'Variables': {
                        'API_NAME': api_name,
                        'S3_BUCKET': s3_bucket_name
                        # 'BENTOML_DEPLOYMENT_PATH_PREFIX': deployment_path_prefix,
                    }
                },
            },
            'Metadata': {
                'Dockertag' : "latest",
                'DockerContext' : project_dir,
                'Dockerfile' : "Dockerfile"
            }
        }

    yaml.dump(sam_config, Path(template_file_path))

    # We add Outputs section separately, because the value should not
    # have "'" around !Sub
    with open(template_file_path, 'a') as f:
        f.write(
            """\
  EndpointUrl:
    Value: !Sub "https://${ServerlessHttpApi}.execute-api.${AWS::Region}.\
amazonaws.com/function/#PROJECT_NAME#"
    Description: URL for endpoint
""".replace("#PROJECT_NAME#",project_name)
        )
    return template_file_path



def _create_aws_lambda_cloudformation_template_file(
    project_dir,
    namespace,
    deployment_name,
    deployment_path_prefix,
    api_names,
    bento_service_name,
    s3_bucket_name,
    py_runtime,
    memory_size,
    timeout,
):
    template_file_path = os.path.join(project_dir, 'template.yaml')
    yaml = YAML()
    sam_config = {
        'AWSTemplateFormatVersion': '2010-09-09',
        'Transform': 'AWS::Serverless-2016-10-31',
        'Globals': {
            'Function': {'Timeout': timeout, 'Runtime': py_runtime},
            'Api': {
                'BinaryMediaTypes': ['image~1*'],
                'Cors': "'*'",
                'Auth': {
                    'ApiKeyRequired': False,
                    'DefaultAuthorizer': 'NONE',
                    'AddDefaultAuthorizerToCorsPreflight': False,
                },
            },
        },
        'Resources': {},
        'Outputs': {
            'S3Bucket': {
                'Value': s3_bucket_name,
                'Description': 'S3 Bucket for saving artifacts and lambda bundle',
            }
        },
    }
    for api_name in api_names:
        sam_config['Resources'][api_name] = {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'Runtime': py_runtime,
                'CodeUri': deployment_name + '/',
                'Handler': 'app.{}'.format(api_name),
                'FunctionName': f'{namespace}-{deployment_name}-{api_name}',
                'Timeout': timeout,
                'MemorySize': memory_size,
                'Events': {
                    'Api': {
                        'Type': 'Api',
                        'Properties': {
                            'Path': '/{}'.format(api_name),
                            'Method': 'post',
                        },
                    }
                },
                'Policies': [{'S3ReadPolicy': {'BucketName': s3_bucket_name}}],
                'Environment': {
                    'Variables': {
                        'BENTOML_API_NAME': api_name,
                        'BENTOML_S3_BUCKET': s3_bucket_name,
                        'BENTOML_DEPLOYMENT_PATH_PREFIX': deployment_path_prefix,
                    }
                },
            },
        }

    yaml.dump(sam_config, Path(template_file_path))

    # We add Outputs section separately, because the value should not
    # have "'" around !Sub
    with open(template_file_path, 'a') as f:
        f.write(
            """\
  EndpointUrl:
    Value: !Sub "https://${ServerlessRestApi}.execute-api.${AWS::Region}.\
amazonaws.com/Prod"
    Description: URL for endpoint
"""
        )
    return template_file_path


# def _deploy_lambda_function(
#     deployment_pb,
#     bento_service_metadata,
#     deployment_spec,
#     lambda_s3_bucket,
#     lambda_deployment_config,
#     bento_path,
# ):
#     deployment_path_prefix = os.path.join(deployment_pb.namespace, deployment_pb.name)

#     py_major, py_minor, _ = bento_service_metadata.env.python_version.split('.')
#     if py_major != '3':
#         raise BentoMLException('Python 2 is not supported for Lambda Deployment')
#     python_runtime = 'python{}.{}'.format(py_major, py_minor)

#     artifact_types = [item.artifact_type for item in bento_service_metadata.artifacts]
#     if any(
#         i in ['TensorflowSavedModelArtifact', 'KerasModelArtifact']
#         for i in artifact_types
#     ) and (py_major, py_minor) != ('3', '6'):
#         raise BentoMLException(
#             'AWS Lambda Deployment only supports BentoML services'
#             'built with Python 3.6.x. To fix this, repack your'
#             'service with the right Python version'
#             '(hint: pyenv/anaconda) and try again'
#         )

#     api_names = (
#         [lambda_deployment_config.api_name]
#         if lambda_deployment_config.api_name
#         else [api.name for api in bento_service_metadata.apis]
#     )

#     raise_if_api_names_not_found_in_bento_service_metadata(
#         bento_service_metadata, api_names
#     )

#     with TempDirectory() as lambda_project_dir:
#         logger.debug(
#             'Generating cloudformation template.yaml for lambda project at %s',
#             lambda_project_dir,
#         )
#         template_file_path = _create_aws_lambda_cloudformation_template_file(
#             project_dir=lambda_project_dir,
#             namespace=deployment_pb.namespace,
#             deployment_name=deployment_pb.name,
#             deployment_path_prefix=deployment_path_prefix,
#             api_names=api_names,
#             bento_service_name=deployment_spec.bento_name,
#             s3_bucket_name=lambda_s3_bucket,
#             py_runtime=python_runtime,
#             memory_size=lambda_deployment_config.memory_size,
#             timeout=lambda_deployment_config.timeout,
#         )
#         logger.debug('Validating generated template.yaml')
#         validate_sam_template(
#             template_file_path, lambda_deployment_config.region, lambda_project_dir,
#         )
#         logger.debug(
#             'Initializing lambda project in directory: %s ...', lambda_project_dir,
#         )
#         init_sam_project(
#             lambda_project_dir,
#             bento_path,
#             deployment_pb.name,
#             deployment_spec.bento_name,
#             api_names,
#             aws_region=lambda_deployment_config.region,
#         )
#         for api_name in api_names:
#             build_directory = os.path.join(
#                 lambda_project_dir, '.aws-sam', 'build', api_name
#             )
#             logger.debug(
#                 'Checking is function "%s" bundle under lambda size ' 'limit', api_name,
#             )
#             # Since we only use s3 get object in lambda function, and
#             # lambda function pack their own boto3/botocore modules,
#             # we will just delete those modules from function bundle
#             # directory
#             delete_list = ['boto3', 'botocore']
#             for name in delete_list:
#                 logger.debug('Remove module "%s" from build directory', name)
#                 shutil.rmtree(os.path.join(build_directory, name))
#             total_build_dir_size = total_file_or_directory_size(build_directory)
#             if total_build_dir_size > LAMBDA_FUNCTION_MAX_LIMIT:
#                 raise BentoMLException(
#                     'Build function size is over 700MB, max size '
#                     'capable for AWS Lambda function'
#                 )
#             if total_build_dir_size >= LAMBDA_FUNCTION_LIMIT:
#                 logger.debug(
#                     'Function %s is over lambda size limit, attempting ' 'reduce it',
#                     api_name,
#                 )
#                 reduce_bundle_size_and_upload_extra_resources_to_s3(
#                     build_directory=build_directory,
#                     region=lambda_deployment_config.region,
#                     s3_bucket=lambda_s3_bucket,
#                     deployment_prefix=deployment_path_prefix,
#                     function_name=api_name,
#                     lambda_project_dir=lambda_project_dir,
#                 )
#             else:
#                 logger.debug(
#                     'Function bundle is within Lambda limit, removing '
#                     'download_extra_resources.py file from function bundle'
#                 )
#                 os.remove(os.path.join(build_directory, 'download_extra_resources.py'))
#         logger.info('Packaging AWS Lambda project at %s ...', lambda_project_dir)
#         lambda_package(
#             lambda_project_dir,
#             lambda_deployment_config.region,
#             lambda_s3_bucket,
#             deployment_path_prefix,
#         )
#         logger.info('Deploying lambda project')
#         stack_name = generate_aws_compatible_string(
#             deployment_pb.namespace + '-' + deployment_pb.name
#         )
#         lambda_deploy(
#             lambda_project_dir, lambda_deployment_config.region, stack_name=stack_name,
#         )


# class AwsLambdaDeploymentOperator(DeploymentOperatorBase):
#     def __init__(self, yatai_service):
#         super(AwsLambdaDeploymentOperator, self).__init__(yatai_service)
#         ensure_docker_available_or_raise()
#         ensure_sam_available_or_raise()

#     def add(self, deployment_pb):
#         try:
#             deployment_spec = deployment_pb.spec
#             deployment_spec.aws_lambda_operator_config.region = (
#                 deployment_spec.aws_lambda_operator_config.region
#                 or get_default_aws_region()
#             )
#             if not deployment_spec.aws_lambda_operator_config.region:
#                 raise InvalidArgument('AWS region is missing')

#             bento_pb = self.yatai_service.GetBento(
#                 GetBentoRequest(
#                     bento_name=deployment_spec.bento_name,
#                     bento_version=deployment_spec.bento_version,
#                 )
#             )
#             if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
#                 raise BentoMLException(
#                     'BentoML currently not support {} repository'.format(
#                         BentoUri.StorageType.Name(bento_pb.bento.uri.type)
#                     )
#                 )

#             return self._add(deployment_pb, bento_pb, bento_pb.bento.uri.uri)
#         except BentoMLException as error:
#             deployment_pb.state.state = DeploymentState.ERROR
#             deployment_pb.state.error_message = f'Error: {str(error)}'
#             return ApplyDeploymentResponse(
#                 status=error.status_proto, deployment=deployment_pb
#             )

#     def _add(self, deployment_pb, bento_pb, bento_path):
#         if loader._is_remote_path(bento_path):
#             with loader._resolve_remote_bundle_path(bento_path) as local_path:
#                 return self._add(deployment_pb, bento_pb, local_path)

#         deployment_spec = deployment_pb.spec
#         lambda_deployment_config = deployment_spec.aws_lambda_operator_config
#         bento_service_metadata = bento_pb.bento.bento_service_metadata
#         lambda_s3_bucket = generate_aws_compatible_string(
#             'btml-{namespace}-{name}-{random_string}'.format(
#                 namespace=deployment_pb.namespace,
#                 name=deployment_pb.name,
#                 random_string=uuid.uuid4().hex[:6].lower(),
#             )
#         )
#         try:
#             create_s3_bucket_if_not_exists(
#                 lambda_s3_bucket, lambda_deployment_config.region
#             )
#             _deploy_lambda_function(
#                 deployment_pb=deployment_pb,
#                 bento_service_metadata=bento_service_metadata,
#                 deployment_spec=deployment_spec,
#                 lambda_s3_bucket=lambda_s3_bucket,
#                 lambda_deployment_config=lambda_deployment_config,
#                 bento_path=bento_path,
#             )
#             return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)
#         except BentoMLException as error:
#             if lambda_s3_bucket and lambda_deployment_config:
#                 cleanup_s3_bucket_if_exist(
#                     lambda_s3_bucket, lambda_deployment_config.region
#                 )
#             raise error

#     def update(self, deployment_pb, previous_deployment):
#         try:
#             deployment_spec = deployment_pb.spec
#             bento_pb = self.yatai_service.GetBento(
#                 GetBentoRequest(
#                     bento_name=deployment_spec.bento_name,
#                     bento_version=deployment_spec.bento_version,
#                 )
#             )
#             if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
#                 raise BentoMLException(
#                     'BentoML currently not support {} repository'.format(
#                         BentoUri.StorageType.Name(bento_pb.bento.uri.type)
#                     )
#                 )

#             return self._update(
#                 deployment_pb, previous_deployment, bento_pb, bento_pb.bento.uri.uri
#             )
#         except BentoMLException as error:
#             deployment_pb.state.state = DeploymentState.ERROR
#             deployment_pb.state.error_message = f'Error: {str(error)}'
#             return ApplyDeploymentResponse(
#                 status=error.status_code, deployment_pb=deployment_pb
#             )

#     def _update(self, deployment_pb, current_deployment, bento_pb, bento_path):
#         if loader._is_remote_path(bento_path):
#             with loader._resolve_remote_bundle_path(bento_path) as local_path:
#                 return self._update(
#                     deployment_pb, current_deployment, bento_pb, local_path
#                 )
#         updated_deployment_spec = deployment_pb.spec
#         updated_lambda_deployment_config = (
#             updated_deployment_spec.aws_lambda_operator_config
#         )
#         updated_bento_service_metadata = bento_pb.bento.bento_service_metadata
#         describe_result = self.describe(deployment_pb)
#         if describe_result.status.status_code != status_pb2.Status.OK:
#             error_code, error_message = status_pb_to_error_code_and_message(
#                 describe_result.status
#             )
#             raise YataiDeploymentException(
#                 f'Failed fetching Lambda deployment current status - '
#                 f'{error_code}:{error_message}'
#             )
#         latest_deployment_state = json.loads(describe_result.state.info_json)
#         if 's3_bucket' in latest_deployment_state:
#             lambda_s3_bucket = latest_deployment_state['s3_bucket']
#         else:
#             raise BentoMLException(
#                 'S3 Bucket is missing in the AWS Lambda deployment, please make sure '
#                 'it exists and try again'
#             )

#         _deploy_lambda_function(
#             deployment_pb=deployment_pb,
#             bento_service_metadata=updated_bento_service_metadata,
#             deployment_spec=updated_deployment_spec,
#             lambda_s3_bucket=lambda_s3_bucket,
#             lambda_deployment_config=updated_lambda_deployment_config,
#             bento_path=bento_path,
#         )

#         return ApplyDeploymentResponse(deployment=deployment_pb, status=Status.OK())

#     def delete(self, deployment_pb):
#         try:
#             logger.debug('Deleting AWS Lambda deployment')

#             deployment_spec = deployment_pb.spec
#             lambda_deployment_config = deployment_spec.aws_lambda_operator_config
#             lambda_deployment_config.region = (
#                 lambda_deployment_config.region or get_default_aws_region()
#             )
#             if not lambda_deployment_config.region:
#                 raise InvalidArgument('AWS region is missing')

#             cf_client = boto3.client('cloudformation', lambda_deployment_config.region)
#             stack_name = generate_aws_compatible_string(
#                 deployment_pb.namespace, deployment_pb.name
#             )
#             if deployment_pb.state.info_json:
#                 deployment_info_json = json.loads(deployment_pb.state.info_json)
#                 bucket_name = deployment_info_json.get('s3_bucket')
#                 if bucket_name:
#                     cleanup_s3_bucket_if_exist(
#                         bucket_name, lambda_deployment_config.region
#                     )

#             logger.debug(
#                 'Deleting AWS CloudFormation: %s that includes Lambda function '
#                 'and related resources',
#                 stack_name,
#             )
#             cf_client.delete_stack(StackName=stack_name)
#             return DeleteDeploymentResponse(status=Status.OK())

#         except BentoMLException as error:
#             return DeleteDeploymentResponse(status=error.status_proto)

#     def describe(self, deployment_pb):
#         try:
#             deployment_spec = deployment_pb.spec
#             lambda_deployment_config = deployment_spec.aws_lambda_operator_config
#             lambda_deployment_config.region = (
#                 lambda_deployment_config.region or get_default_aws_region()
#             )
#             if not lambda_deployment_config.region:
#                 raise InvalidArgument('AWS region is missing')

#             bento_pb = self.yatai_service.GetBento(
#                 GetBentoRequest(
#                     bento_name=deployment_spec.bento_name,
#                     bento_version=deployment_spec.bento_version,
#                 )
#             )
#             bento_service_metadata = bento_pb.bento.bento_service_metadata
#             api_names = (
#                 [lambda_deployment_config.api_name]
#                 if lambda_deployment_config.api_name
#                 else [api.name for api in bento_service_metadata.apis]
#             )

#             try:
#                 cf_client = boto3.client(
#                     'cloudformation', lambda_deployment_config.region
#                 )
#                 stack_name = generate_aws_compatible_string(
#                     '{ns}-{name}'.format(
#                         ns=deployment_pb.namespace, name=deployment_pb.name
#                     )
#                 )
#                 cloud_formation_stack_result = cf_client.describe_stacks(
#                     StackName=stack_name
#                 )
#                 stack_result = cloud_formation_stack_result.get('Stacks')[0]
#                 # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/\
#                 # using-cfn-describing-stacks.html
#                 success_status = ['CREATE_COMPLETE', 'UPDATE_COMPLETE']
#                 if stack_result['StackStatus'] in success_status:
#                     if stack_result.get('Outputs'):
#                         outputs = stack_result['Outputs']
#                     else:
#                         return DescribeDeploymentResponse(
#                             status=Status.ABORTED('"Outputs" field is not present'),
#                             state=DeploymentState(
#                                 state=DeploymentState.ERROR,
#                                 error_message='"Outputs" field is not present',
#                             ),
#                         )
#                 elif stack_result['StackStatus'] in FAILED_CLOUDFORMATION_STACK_STATUS:
#                     state = DeploymentState(state=DeploymentState.FAILED)
#                     state.timestamp.GetCurrentTime()
#                     return DescribeDeploymentResponse(status=Status.OK(), state=state)
#                 else:
#                     state = DeploymentState(state=DeploymentState.PENDING)
#                     state.timestamp.GetCurrentTime()
#                     return DescribeDeploymentResponse(status=Status.OK(), state=state)
#             except Exception as error:  # pylint: disable=broad-except
#                 state = DeploymentState(
#                     state=DeploymentState.ERROR, error_message=str(error)
#                 )
#                 state.timestamp.GetCurrentTime()
#                 return DescribeDeploymentResponse(
#                     status=Status.INTERNAL(str(error)), state=state
#                 )
#             outputs = {o['OutputKey']: o['OutputValue'] for o in outputs}
#             info_json = {}

#             if 'EndpointUrl' in outputs:
#                 info_json['endpoints'] = [
#                     outputs['EndpointUrl'] + '/' + api_name for api_name in api_names
#                 ]
#             if 'S3Bucket' in outputs:
#                 info_json['s3_bucket'] = outputs['S3Bucket']

#             state = DeploymentState(
#                 state=DeploymentState.RUNNING, info_json=json.dumps(info_json)
#             )
#             state.timestamp.GetCurrentTime()
#             return DescribeDeploymentResponse(status=Status.OK(), state=state)
#         except BentoMLException as error:
#             return DescribeDeploymentResponse(status=error.status_proto)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import logging
import re
import tarfile
from pathlib import Path

# import boto3
# from botocore.exceptions import ClientError

# from bentoml.exceptions import BentoMLException
# from bentoml.yatai.deployment.aws_utils import call_sam_command
import subprocess

logger = logging.getLogger(__name__)

# Maximum size for Lambda function bundle 250MB, leaving 1mb offset
LAMBDA_FUNCTION_LIMIT = 249000000
# Total disk size that user has access to.  Lambda function bundle + /tmp 512MB, leaving
# 1mb offset
LAMBDA_TEMPORARY_DIRECTORY_MAX_LIMIT = 511000000
LAMBDA_FUNCTION_MAX_LIMIT = LAMBDA_FUNCTION_LIMIT + LAMBDA_TEMPORARY_DIRECTORY_MAX_LIMIT

def call_sam_command(command, project_dir, region):
    command = ["sam"] + command

    # We are passing region as part of the param, due to sam cli is not currently
    # using the region that passed in each command.  Set the region param as
    # AWS_DEFAULT_REGION for the subprocess call
    logger.debug('Setting envar "AWS_DEFAULT_REGION" to %s for subprocess call', region)
    copied_env = os.environ.copy()
    copied_env["AWS_DEFAULT_REGION"] = region

    proc = subprocess.Popen(
        command,
        cwd=project_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=copied_env,
    )
    stdout, stderr = proc.communicate()
    logger.debug("SAM cmd %s output: %s", command, stdout.decode("utf-8"))
    return proc.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")

def call_bash(command, project_dir):
    proc = subprocess.Popen(
        command,
        cwd=project_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = proc.communicate()
    # logger.debug("SAM cmd %s output: %s", command, stdout.decode("utf-8"))
    return proc.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")

# def init_sam_project(
#     sam_project_path,
#     bento_service_bundle_path,
#     deployment_name,
#     bento_name,
#     api_names,
#     aws_region,
# ):
#     function_path = os.path.join(sam_project_path, deployment_name)
#     os.mkdir(function_path)
#     # Copy requirements.txt
#     logger.debug("Copying requirements.txt")
#     requirement_txt_path = os.path.join(bento_service_bundle_path, "requirements.txt")
#     shutil.copy(requirement_txt_path, function_path)

#     bundled_dep_path = os.path.join(
#         bento_service_bundle_path, "bundled_pip_dependencies"
#     )
#     if os.path.isdir(bundled_dep_path):
#         # Copy bundled pip dependencies
#         logger.debug("Copying bundled_dependencies")
#         shutil.copytree(
#             bundled_dep_path, os.path.join(function_path, "bundled_pip_dependencies")
#         )
#         bundled_files = os.listdir(
#             os.path.join(function_path, "bundled_pip_dependencies")
#         )
#         for index, bundled_file_name in enumerate(bundled_files):
#             bundled_files[index] = "./bundled_pip_dependencies/{}\n".format(
#                 bundled_file_name
#             )

#         logger.debug("Updating requirements.txt")
#         with open(
#             os.path.join(function_path, "requirements.txt"), "r+"
#         ) as requirement_file:
#             required_modules = bundled_files + requirement_file.readlines()
#             # Write from beginning of the file, instead of appending to the end.
#             requirement_file.seek(0)
#             requirement_file.writelines(required_modules)

#     # Copy bento_service_model
#     logger.debug("Copying model directory")
#     model_path = os.path.join(bento_service_bundle_path, bento_name)
#     shutil.copytree(model_path, os.path.join(function_path, bento_name))

#     logger.debug("Creating python files for lambda function")
#     # Create empty __init__.py
#     open(os.path.join(function_path, "__init__.py"), "w").close()

#     app_py_path = os.path.join(os.path.dirname(__file__), "lambda_app.py")
#     shutil.copy(app_py_path, os.path.join(function_path, "app.py"))
#     unzip_requirement_py_path = os.path.join(
#         os.path.dirname(__file__), "download_extra_resources.py"
#     )
#     shutil.copy(
#         unzip_requirement_py_path,
#         os.path.join(function_path, "download_extra_resources.py"),
#     )

#     logger.info("Building lambda project")
#     return_code, stdout, stderr = call_sam_command(
#         ["build", "--use-container", "--region", aws_region],
#         project_dir=sam_project_path,
#         region=aws_region,
#     )
#     if return_code != 0:
#         error_message = stderr
#         if not error_message:
#             error_message = stdout
#         raise BentoMLException(
#             "Failed to build lambda function. {}".format(error_message)
#         )
#     logger.debug("Removing unnecessary files to free up space")
#     for api_name in api_names:
#         cleanup_build_files(sam_project_path, api_name)


def cleanup_build_files(project_dir, api_name):
    build_dir = os.path.join(project_dir, ".aws-sam/build/{}".format(api_name))
    logger.debug("Cleaning up unused files in SAM built directory %s", build_dir)
    for root, dirs, files in os.walk(build_dir):
        if "tests" in dirs:
            logger.debug("removing dir: %s", os.path.join(root, "tests"))
            shutil.rmtree(os.path.join(root, "tests"))
        if "__pycache__" in dirs:
            logger.debug("removing dir: %s", os.path.join(root, "__pycache__"))
            shutil.rmtree(os.path.join(root, "__pycache__"))

        for dir_name in filter(lambda x: re.match("^.*\\.dist-info$", x), dirs):
            logger.debug("removing dir: %s", dir_name)
            shutil.rmtree(os.path.join(root, dir_name))

        for file in filter(lambda x: re.match("^.*\\.egg-info$", x), files):
            logger.debug("removing file: %s", os.path.join(root, file))
            os.remove(os.path.join(root, file))

        for file in filter(lambda x: re.match("^.*\\.pyc$", x), files):
            logger.debug("removing file: %s", os.path.join(root, file))
            os.remove(os.path.join(root, file))

        if "wheel" in files:
            logger.debug("removing file: %s", os.path.join(root, "wheel"))
            os.remove(os.path.join(root, "wheel"))
        if "pip" in files:
            logger.debug("removing file: %s", os.path.join(root, "pip"))
            os.remove(os.path.join(root, "pip"))
        if "caff2" in files:
            logger.debug("removing file: %s", os.path.join(root, "caff2"))
            os.remove(os.path.join(root, "caff2"))


# def lambda_package(project_dir, aws_region, s3_bucket_name, deployment_prefix):
#     prefix_path = os.path.join(deployment_prefix, "lambda-functions")
#     build_dir = os.path.join(project_dir, ".aws-sam", "build")

#     return_code, stdout, stderr = call_sam_command(
#         [
#             "package",
#             "--force-upload",
#             "--s3-bucket",
#             s3_bucket_name,
#             "--s3-prefix",
#             prefix_path,
#             "--template-file",
#             "template.yaml",
#             "--output-template-file",
#             "packaged.yaml",
#             "--region",
#             aws_region,
#         ],
#         project_dir=build_dir,
#         region=aws_region,
#     )
#     if return_code != 0:
#         error_message = stderr
#         if not error_message:
#             error_message = stdout
#         raise BentoMLException(
#             "Failed to package lambda function. {}".format(error_message)
#         )
#     else:
#         return stdout


def lambda_deploy(project_dir, aws_region, stack_name):
    # if the stack name exists and the state is in rollback_complete or
    # other 'bad' state, we will delete the stack first, and then deploy
    # it
    logger.debug('Ensure stack "%s" is ready to deploy', stack_name)
    ensure_is_ready_to_deploy_to_cloud_formation(stack_name, aws_region)
    logger.debug('Stack "%s"is ready to deploy', stack_name)

    template_file = os.path.join(project_dir, ".aws-sam", "build", "packaged.yaml")
    return_code, stdout, stderr = call_sam_command(
        [
            "deploy",
            "--stack-name",
            stack_name,
            "--capabilities",
            "CAPABILITY_IAM",
            "--template-file",
            template_file,
            "--region",
            aws_region,
        ],
        project_dir=project_dir,
        region=aws_region,
    )
    if return_code != 0:
        error_message = stderr
        if not error_message:
            error_message = stdout
        raise BentoMLException(
            "Failed to deploy lambda function. {}".format(error_message)
        )
    else:
        return stdout


# def total_file_or_directory_size(file_or_dir):
#     if os.path.isdir(file_or_dir):
#         return sum(
#             f.stat().st_size for f in Path(file_or_dir).glob("**/*") if f.is_file()
#         )
#     else:
#         return Path(file_or_dir).stat().st_size


def check_aws_cli_exists():
    return_code, stdout, stderr = call_bash(["aws"],os.getcwd())

    # print(return_code)
    # print(stdout)
    # print(stderr)

    if "not found" in stdout or "not found" in stderr:
        print("Lambda deployment needs aws cli to be installed : https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-install.html")
        exit(0)

    # return_code, stdout, stderr = call_bash["aws","sts","get-caller-identity"]

    # if "Unable to locate credentials" in stdout or "Unable to locate credentials" in stderr:
    #     print("aws not configured : aws configure")
    #     exit(0)

def check_sam_cli_exists():
    eturn_code, stdout, stderr = call_bash(["sam"],os.getcwd())

    # print(return_code)
    # print(stdout)
    # print(stderr)

    if "not found" in stdout or "not found" in stderr:
        print("Lambda deployment needs sam cli to be installed : https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-install.html")
        exit(0)

# def reduce_bundle_size_and_upload_extra_resources_to_s3(
#     build_directory,
#     region,
#     s3_bucket,
#     deployment_prefix,
#     function_name,
#     lambda_project_dir,
# ):
#     additional_requirements_path = os.path.join(
#         lambda_project_dir, "additional_requirements", function_name
#     )
#     upload_dir_path = os.path.join(additional_requirements_path, "requirements")
#     os.makedirs(upload_dir_path, exist_ok=True)
#     s3_client = boto3.client("s3", region)

#     dir_name_to_size = dict(
#         (item, total_file_or_directory_size(os.path.join(build_directory, item)))
#         for item in os.listdir(build_directory)
#     )

#     required_bundle_list = ["app.py", "__init__.py", "download_extra_resources.py"]
#     required_bundle_size = sum(dir_name_to_size[i] for i in required_bundle_list)
#     for name, size in sorted(
#         dir_name_to_size.items(), key=lambda i: i[1], reverse=True
#     ):
#         if name in required_bundle_list:
#             logger.debug('Including "%s" it is part of required bundle.', name)
#             continue
#         else:
#             new_bundle_size = required_bundle_size + size
#             if new_bundle_size >= LAMBDA_FUNCTION_LIMIT:
#                 logger.debug(
#                     'Lambda function size %d is over the limit. Moving item "%s" '
#                     "out of the bundle directory",
#                     new_bundle_size,
#                     name,
#                 )
#                 package_path = os.path.join(build_directory, name)
#                 copy_dst_path = os.path.join(upload_dir_path, name)
#                 shutil.move(package_path, copy_dst_path)
#             else:
#                 logger.debug(
#                     'Including "%s" Now the current lambda function size is %d',
#                     name,
#                     new_bundle_size,
#                 )
#                 required_bundle_size = new_bundle_size

#     logger.debug("Final Lambda function bundle size is %d", required_bundle_size)
#     additional_requirements_size = total_file_or_directory_size(
#         additional_requirements_path
#     )
#     logger.debug("Additional requirement size is %d", additional_requirements_size)
#     logger.debug("zip up additional requirement packages")
#     tar_file_path = os.path.join(additional_requirements_path, "requirements.tar")
#     with tarfile.open(tar_file_path, "w:gz") as tar:
#         tar.add(upload_dir_path, arcname="requirements")
#     logger.debug("Uploading requirements.tar to %s/%s", s3_bucket, deployment_prefix)
#     try:
#         s3_client.upload_file(
#             tar_file_path,
#             s3_bucket,
#             os.path.join(deployment_prefix, "requirements.tar"),
#         )
#     except ClientError as e:
#         raise BentoMLException(str(e))


# def ensure_is_ready_to_deploy_to_cloud_formation(stack_name, region):
#     try:
#         cf_client = boto3.client("cloudformation", region)
#         logger.debug("Checking stack description")
#         describe_formation_result = cf_client.describe_stacks(StackName=stack_name)
#         result_stacks = describe_formation_result.get("Stacks")
#         if len(result_stacks):
#             logger.debug('Stack "%s" exists', stack_name)
#             stack_result = result_stacks[0]
#             if stack_result["StackStatus"] in [
#                 "ROLLBACK_COMPLETE",
#                 "ROLLBACK_FAILED",
#                 "ROLLBACK_IN_PROGRESS",
#             ]:
#                 logger.debug(
#                     'Stack "%s" is in a "bad" status(%s), deleting the stack '
#                     "before deployment",
#                     stack_name,
#                     stack_result["StackStatus"],
#                 )
#                 cf_client.delete_stack(StackName=stack_name)
#     except ClientError as e:
#         # We are brutally parse and handle stack doesn't exist, since
#         # "AmazonCloudFormationException" currently is not implemented in boto3. Once
#         # the current error is implemented, we need to switch
#         error_response = e.response.get("Error", {})
#         error_code = error_response.get("Code")
#         error_message = error_response.get("Message", "Unknown")
#         if error_code == "ValidationError" and "does not exist" in error_message:
#             pass
#         else:
#             raise BentoMLException(str(e))





# from clouderizer_beta.operator_this import AwsLambdaDeploymentOperator
from clouderizer_beta.operator_this import create_aws_lambda_cloudformation_template_file
from clouderizer_beta.utils import call_sam_command,call_bash,cleanup_build_files
import shutil
import os
from pathlib import Path
import json
import site

# def create(self, deployment_info, wait):
# 		if isinstance(deployment_info, dict):
# 			deployment_pb = deployment_dict_to_pb(deployment_info)
# 		elif isinstance(deployment_info, str):
# 			deployment_pb = deployment_yaml_string_to_pb(deployment_info)
# 		elif isinstance(deployment_info, Deployment):
# 			deployment_pb = deployment_info
# 		else:
# 			raise YataiDeploymentException(
# 				'Unexpected argument type, expect deployment info to be str in yaml '
# 				'format or a dict or a deployment protobuf obj, instead got: {}'.format(
# 					str(type(deployment_info))
# 				)
# 			)

# 		validation_errors = validate_deployment_pb(deployment_pb)
# 		if validation_errors:
# 			raise YataiDeploymentException(
# 				f'Failed to validate deployment {deployment_pb.name}: '
# 				f'{validation_errors}'
# 			)
# 		# Make sure there is no active deployment with the same deployment name
# 		get_deployment_pb = self.yatai_service.GetDeployment(
# 			GetDeploymentRequest(
# 				deployment_name=deployment_pb.name, namespace=deployment_pb.namespace
# 			)
# 		)
# 		if get_deployment_pb.status.status_code != status_pb2.Status.NOT_FOUND:
# 			raise BentoMLException(
# 				f'Deployment "{deployment_pb.name}" already existed, use Update or '
# 				f'Apply for updating existing deployment, delete the deployment, '
# 				f'or use a different deployment name'
# 			)
# 		apply_result = self.yatai_service.ApplyDeployment(
# 			ApplyDeploymentRequest(deployment=deployment_pb)
# 		)
# 		if apply_result.status.status_code != status_pb2.Status.OK:
# 			error_code, error_message = status_pb_to_error_code_and_message(
# 				apply_result.status
# 			)
# 			raise YataiDeploymentException(f'{error_code}:{error_message}')
# 		if wait:
# 			self._wait_deployment_action_complete(
# 				deployment_pb.name, deployment_pb.namespace
# 			)
# 		return self.get(namespace=deployment_pb.namespace, name=deployment_pb.name)

# def get(self, namespace, name):
# 		return self.yatai_service.GetDeployment(
# 			GetDeploymentRequest(deployment_name=name, namespace=namespace)
# 		)
# print(os.getcwd())

def lambda_deploy(base_dir,namespace,project_name,stackName, projectConfig, model_path,infra,pipPackages):

		clouderizerProjectName = project_name
		project_name=project_name.replace("-","")
		stackName=stackName.replace("-","")
		print("Subtype lambda_deploy type:", projectConfig["subtype"])

		project_dir = base_dir + "/lambda-dir"
		Path(project_dir).mkdir(parents=True, exist_ok=True)

		function_dir = project_dir + "/function"
		Path(function_dir).mkdir(parents=True, exist_ok=True)



		if projectConfig["subtype"] in ["h2o","pmml","dai","pmml4s"]:
			# shutil.copy(base_dir+'/dockerfiles/AutomlDockerfile',project_dir+"/Dockerfile")
			print("PROJECT DIR",project_dir)
			shutil.copy(site.getusersitepackages()+'/clouderizer_beta/dockerfiles/AutomlDockerfile',project_dir+"/Dockerfile")

		if projectConfig["subtype"] in ["python","pythonscore","onnx","train"]:
			# shutil.copy(base_dir+'/dockerfiles/PythonDockerfile',project_dir+"/Dockerfile")
			shutil.copy(site.getusersitepackages()+'/clouderizer_beta/dockerfiles/PythonDockerfile',project_dir+"/Dockerfile")
			# os.mknod(function_dir+"/requirements.txt")
			Path(function_dir+"/requirements.txt").open(mode='w')

			pipPackagesFile = open(function_dir+"/requirements.txt",'w')
			print("pipPackages here",pipPackages)
			if len(pipPackages)>0:
				for pipPackage in pipPackages:
					# pipPackagesFile.write("\n")
					pipPackagesFile.write(pipPackage+"\n")

			pipPackagesFile.close()

		writeObj = open(function_dir+"/projectConfig.json","w")
		writeObj.write(json.dumps(projectConfig))
		writeObj.close()

		shutil.copy(model_path,function_dir+"/model.pmml")

		apiNames = [stackName+"Api"]
		stack_name = stackName

		ecr_name = "clouderizer-lambda"

		print("Checking for default aws region configured in aws cli")

		awsRegionCode, awsRegionStdout, awsRegionStderr = call_bash(
			["aws","configure","get","region"],
			project_dir
		)

		awsRegionStdout = awsRegionStdout.replace("\n","")
		print("Configured region: ", awsRegionStdout)

		aws_region = awsRegionStdout
		if not awsRegionStdout:
			aws_region = "us-east-2"
			print("Choosing ",aws_region)

		return_code, stdout, stderr = call_bash(
			["aws", "ecr", "create-repository", "--repository-name" ,ecr_name,
			"--image-tag-mutability", "IMMUTABLE", "--image-scanning-configuration" ,"scanOnPush=true"],
			project_dir
		)

		# if not "RepositoryAlreadyExistsException" in stderr:
		return_code, stdout, stderr = call_bash(
			["aws", "sts", "get-caller-identity"],
			project_dir
		)

		# print(eval(stdout)["Account"])

		if not "Account" in eval(stdout):
			print("Could not fetch aws account")
			exit(0)

		registry_url = "{}.dkr.ecr.{}.amazonaws.com/{}".format(eval(stdout)["Account"],aws_region,ecr_name)

		if infra == "standard":
			memory = 2000

		if infra == "highmemory":
			memory = 8000

		create_aws_lambda_cloudformation_template_file(
			clouderizerProjectName,
			project_dir,
			namespace,
			project_name,
			apiNames,
			project_name,
			memory,
			300
		)

		shutil.copy(project_dir+'/template.yaml',os.getcwd()+'/template.yaml')

		print("Building docker image....")

		return_code, stdout, stderr = call_sam_command(
			["build", "--use-container", "--region", aws_region],
			project_dir=os.getcwd(),
			region=aws_region,
		)
		if return_code != 0:
			error_message = stderr

			if error_message:
				print(error_message)
				exit(0)

			# print(stderr)
			# print(stdout)
			# raise BentoMLException(
			#     "Failed to build lambda function. {}".format(error_message)
			# )
		# logger.debug("Removing unnecessary files to free up space")
		template_file = os.path.join(os.getcwd(), ".aws-sam", "build", "template.yaml")

		print("Deploying project to AWS lambda")

		return_code, stdout, stderr = call_sam_command(
			[
				"deploy",
				"--stack-name",
				stack_name,
				"--capabilities",
				"CAPABILITY_IAM",
				"--template-file",
				template_file,
				"--region",
				aws_region,
				"--image-repositories",
				"{}={}".format(apiNames[0],registry_url)
			],
			project_dir=project_dir,
			region=aws_region,
    	)
		if return_code != 0:
			error_message = stderr
			if error_message:
				print(error_message)
				exit(0)

		print(stdout)
		print("Project deployed!")
		print("Fetching project access urls")

		describeStackCode, describeStackStdout, describeStackStderr = call_bash(
			[
				"aws",
				"cloudformation",
				"describe-stacks"
			],
			project_dir
		)

		accessUrl=None
		if describeStackCode == 0:
			describeStackStdout = json.loads(describeStackStdout)
			for stack in describeStackStdout["Stacks"]:
				if stack["StackName"] == stackName:
					for output in stack["Outputs"]:
						if output["OutputKey"] == "EndpointUrl":
							accessUrl = output["OutputValue"]
			print("Model access url:", accessUrl)
			return accessUrl
		else:
			print(describeStackStderr)
			exit(0)

		cleanup_build_files(project_dir,apiNames[0])

		
